from django.apps import AppConfig


class PostgresExtraAppConfig(AppConfig):
    name = "psqlextra"
    verbose_name = "PostgreSQL Extra"

    def ready(self) -> None:
        from .lookups import InValuesLookup  # noqa



import importlib

from django.conf import settings
from django.core.exceptions import ImproperlyConfigured
from django.db import DEFAULT_DB_ALIAS, connections

from django.db.backends.postgresql.base import (  # isort:skip
    DatabaseWrapper as Psycopg2DatabaseWrapper,
)


def base_backend_instance():
    """Gets an instance of the base class for the custom database back-end.

    This should be the Django PostgreSQL back-end. However,
    some people are already using a custom back-end from
    another package. We are nice people and expose an option
    that allows them to configure the back-end we base upon.

    As long as the specified base eventually also has
    the PostgreSQL back-end as a base, then everything should
    work as intended.

    We create an instance to inspect what classes to subclass
    because not all back-ends set properties such as `ops_class`
    properly. The PostGIS back-end is a good example.
    """
    base_class_name = getattr(
        settings,
        "POSTGRES_EXTRA_DB_BACKEND_BASE",
        "django.db.backends.postgresql",
    )

    base_class_module = importlib.import_module(base_class_name + ".base")
    base_class = getattr(base_class_module, "DatabaseWrapper", None)

    if not base_class:
        raise ImproperlyConfigured(
            (
                "'%s' is not a valid database back-end."
                " The module does not define a DatabaseWrapper class."
                " Check the value of POSTGRES_EXTRA_DB_BACKEND_BASE."
            )
            % base_class_name
        )

    if isinstance(base_class, Psycopg2DatabaseWrapper):
        raise ImproperlyConfigured(
            (
                "'%s' is not a valid database back-end."
                " It does inherit from the PostgreSQL back-end."
                " Check the value of POSTGRES_EXTRA_DB_BACKEND_BASE."
            )
            % base_class_name
        )

    base_instance = base_class(connections.databases[DEFAULT_DB_ALIAS])
    if base_instance.connection:
        raise ImproperlyConfigured(
            (
                "'%s' establishes a connection during initialization."
                " This is not expected and can lead to more connections"
                " being established than neccesarry."
            )
            % base_class_name
        )

    return base_instance


def backend():
    """Gets the base class for the database back-end."""

    return base_backend_instance().__class__


def schema_editor():
    """Gets the base class for the schema editor.

    We have to use the configured base back-end's schema editor for
    this.
    """

    return base_backend_instance().SchemaEditorClass


def introspection():
    """Gets the base class for the introspection class.

    We have to use the configured base back-end's introspection class
    for this.
    """

    return base_backend_instance().introspection.__class__


def operations():
    """Gets the base class for the operations class.

    We have to use the configured base back-end's operations class for
    this.
    """

    return base_backend_instance().ops.__class__



from psqlextra.backend.migrations.state import PostgresHashPartitionState

from .partition import PostgresPartitionOperation


class PostgresAddHashPartition(PostgresPartitionOperation):
    """Adds a new hash partition to a :see:PartitionedPostgresModel.

    Each partition will hold the rows for which the hash value of the
    partition key divided by the specified modulus will produce the
    specified remainder.
    """

    def __init__(
        self, model_name: str, name: str, modulus: int, remainder: int
    ):
        """Initializes new instance of :see:AddHashPartition.
        Arguments:
            model_name:
                The name of the :see:PartitionedPostgresModel.

            name:
                The name to give to the new partition table.

            modulus:
                Integer value by which the key is divided.

            remainder:
                The remainder of the hash value when divided by modulus.
        """

        super().__init__(model_name, name)

        self.modulus = modulus
        self.remainder = remainder

    def state_forwards(self, app_label, state):
        model = state.models[(app_label, self.model_name_lower)]
        model.add_partition(
            PostgresHashPartitionState(
                app_label=app_label,
                model_name=self.model_name,
                name=self.name,
                modulus=self.modulus,
                remainder=self.remainder,
            )
        )

        state.reload_model(app_label, self.model_name_lower)

    def database_forwards(self, app_label, schema_editor, from_state, to_state):
        model = to_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.add_hash_partition(
                model, self.name, self.modulus, self.remainder
            )

    def database_backwards(
        self, app_label, schema_editor, from_state, to_state
    ):
        model = from_state.apps.get_model(app_label, self.model_name)
        if self.allow_migrate_model(schema_editor.connection.alias, model):
            schema_editor.delete_partition(model, self.name)

    def deconstruct(self):
        name, args, kwargs = super().deconstruct()

        kwargs["modulus"] = self.modulus
        kwargs["remainder"] = self.remainder

        return name, args, kwargs

    def describe(self) -> str:
        return "Creates hash partition %s on %s" % (self.name, self.model_name)



from .case_insensitive_unique_index import CaseInsensitiveUniqueIndex
from .conditional_unique_index import ConditionalUniqueIndex
from .unique_index import UniqueIndex

__all__ = [
    "UniqueIndex",
    "ConditionalUniqueIndex",
    "CaseInsensitiveUniqueIndex",
]



from psqlextra.compiler import (
    SQLAggregateCompiler,
    SQLCompiler,
    SQLDeleteCompiler,
    SQLInsertCompiler,
    SQLUpdateCompiler,
)

from . import base_impl


class PostgresOperations(base_impl.operations()):
    """Simple operations specific to PostgreSQL."""

    compiler_module = "psqlextra.compiler"

    compiler_classes = [
        SQLCompiler,
        SQLDeleteCompiler,
        SQLAggregateCompiler,
        SQLUpdateCompiler,
        SQLInsertCompiler,
    ]



from django.db.models import lookups
from django.db.models.fields import Field, related_lookups
from django.db.models.fields.related import ForeignObject


class InValuesLookupMixin:
    """Performs a `lhs IN VALUES ((a), (b), (c))` lookup.

    This can be significantly faster then a normal `IN (a, b, c)`. The
    latter sometimes causes the Postgres query planner do a sequential
    scan.
    """

    def as_sql(self, compiler, connection):

        if not self.rhs_is_direct_value():
            return super().as_sql(compiler, connection)

        lhs, lhs_params = self.process_lhs(compiler, connection)

        _, rhs_params = self.process_rhs(compiler, connection)
        rhs = ",".join([f"(%s)" for _ in rhs_params])  # noqa: F541

        return f"{lhs} IN (VALUES {rhs})", lhs_params + list(rhs_params)


@Field.register_lookup
class InValuesLookup(InValuesLookupMixin, lookups.In):
    lookup_name = "invalues"


@ForeignObject.register_lookup
class InValuesRelatedLookup(InValuesLookupMixin, related_lookups.RelatedIn):
    lookup_name = "invalues"



import django

from django.db.models.indexes import Index


class UniqueIndex(Index):
    def create_sql(self, *args, **kwargs):
        if django.VERSION >= (2, 0):
            statement = super().create_sql(*args, **kwargs)
            statement.template = self._rewrite_sql(statement.template)
            return statement

        sql = super().create_sql(*args, **kwargs)
        return self._rewrite_sql(sql)

    @staticmethod
    def _rewrite_sql(sql: str) -> str:
        return sql.replace("CREATE INDEX", "CREATE UNIQUE INDEX")



import logging

from django.conf import settings
from django.db import ProgrammingError

from . import base_impl
from .introspection import PostgresIntrospection
from .operations import PostgresOperations
from .schema import PostgresSchemaEditor

logger = logging.getLogger(__name__)


class DatabaseWrapper(base_impl.backend()):
    """Wraps the standard PostgreSQL database back-end.

    Overrides the schema editor with our custom schema editor and makes
    sure the `hstore` extension is enabled.
    """

    SchemaEditorClass = PostgresSchemaEditor
    introspection_class = PostgresIntrospection
    ops_class = PostgresOperations

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # Some base back-ends such as the PostGIS back-end don't properly
        # set `ops_class` and `introspection_class` and initialize these
        # classes themselves.
        #
        # This can lead to broken functionality. We fix this automatically.

        if not isinstance(self.introspection, self.introspection_class):
            self.introspection = self.introspection_class(self)

        if not isinstance(self.ops, self.ops_class):
            self.ops = self.ops_class(self)

        for expected_compiler_class in self.ops.compiler_classes:
            compiler_class = self.ops.compiler(expected_compiler_class.__name__)

            if not issubclass(compiler_class, expected_compiler_class):
                logger.warning(
                    "Compiler '%s.%s' is not properly deriving from '%s.%s'."
                    % (
                        compiler_class.__module__,
                        compiler_class.__name__,
                        expected_compiler_class.__module__,
                        expected_compiler_class.__name__,
                    )
                )

    def prepare_database(self):
        """Ran to prepare the configured database.

        This is where we enable the `hstore` extension if it wasn't
        enabled yet.
        """

        super().prepare_database()

        setup_ext = getattr(
            settings, "POSTGRES_EXTRA_AUTO_EXTENSION_SET_UP", True
        )
        if not setup_ext:
            return False

        with self.cursor() as cursor:
            try:
                cursor.execute("CREATE EXTENSION IF NOT EXISTS hstore")
            except ProgrammingError:  # permission denied
                logger.warning(
                    'Failed to create "hstore" extension. '
                    "Tables with hstore columns may fail to migrate. "
                    "If hstore is needed, make sure you are connected "
                    "to the database as a superuser "
                    "or add the extension manually.",
                    exc_info=True,
                )
